import os
import json
import random
import time
import subprocess
import uuid

from policy_service_integration.api_client.policyservice import PolicyApi
from python_sensor_sim import PythonSensorSim
from python_sensor_sim.enums import SensorDeviceType
from psc_server_support.api.api_factory import ApiFactory
from psc_server_support.api.front_end_authentication import FrontEndAuthentication

CSR_HOST = "https://csr-dev01.defense-dev.carbonblack.io"
DEVICES_URL = os.getenv("DEVICES_URL", "https://dev-dev01.cbdtest.io")
DEFENSE_URL = os.getenv("DEFENSE_URL", "https://defense-dev01.cbdtest.io")
USERNAME = "ronkar+pas-e2e.org@vmware.com"
PASSWORD = os.getenv("DEV01_PAS_E2E_PASSWORD", "")
ENV = "dev01"

SENSOR_GROUPS_UPGRADE_USERNAME = "ronkar+pas-e2e-sensor-upgrade.org@vmware.com"
SENSOR_GROUPS_UPGRADE_PASSWORD = os.getenv("DEV01_PAS_E2E_SENSOR_GROUPS_UPGRADE_PASSWORD", "")

LOCAL_PAS_HOST = "http://policy-assignment-service:8080/policy-assignment"

DEVICES = []
ASSET_GROUPS = []
POLICIES = []

EXPECTED_CALL_TO_SUCCEED = "Expected call to succeed."

TIME = None

ENDPOINTS = []


def setup_data(session):
    global TIME, ENDPOINTS
    TIME = int(time.time())
    ENDPOINTS.extend(search_endpoints(session).get("results"))
    ASSET_GROUPS.extend(search_asset_groups(session).get("results"))


def cleanup(session):
    delete_sensors(session)
    delete_asset_groups(session)
    delete_policies(session)


def sim_endpoints(session, suffix=""):
    api_user = ApiFactory(FrontEndAuthentication.map_auth_session_to_front_end_auth(session))
    for _ in range(1, 9):
        sim_endpoint(session, api_user, suffix=f"{_}-{TIME}" if suffix is None else suffix)


def sim_endpoint(session, api_user=None, suffix="", device_type=None):
    if not api_user:
        api_user = ApiFactory(FrontEndAuthentication.map_auth_session_to_front_end_auth(session))

    if device_type is None:
        device_type = random.choice([SensorDeviceType.WINDOWS, SensorDeviceType.LINUX, SensorDeviceType.MAC])

    sensor_name = f'pas-e2e-{suffix}'
    sensor_sim = PythonSensorSim(
        url=DEVICES_URL,
        device_type=device_type,
        org_user=api_user,
        org_id=session.get_user_org_id(),
        org_key=session.get_user_org_key(),
        device_name=sensor_name)

    DEVICES.append(sensor_sim.device_info)
    return sensor_sim.device_info


def setup_policies(session):
    policy_api = PolicyApi(session)
    policies_summary = get_policies_summary(session=session, policy_api=policy_api)
    base_policy_id = policies_summary.get("policies")[0].get("id")

    response = policy_api.get_policy_details(org_key=session.get_user_org_key(), policy_id=base_policy_id)
    assert response.ok, EXPECTED_CALL_TO_SUCCEED
    policy_details = json.loads(response.content)

    for _ in range(1, 4):
        policy_details['name'] = f'Policy-{_}-{TIME}'
        policy_details['id'] = None
        policy_details['description'] = "Test policy created for PAS tests."

        create_policy_response = policy_api.create_policy(org_key=session.get_user_org_key(), body=policy_details)
        assert create_policy_response.ok, EXPECTED_CALL_TO_SUCCEED

        POLICIES.append(json.loads(create_policy_response.content))


def get_policies_summary(session, policy_api=None):
    if policy_api is None:
        policy_api = PolicyApi(session)

    response = policy_api.get_policies_summary(org_key=session.get_user_org_key())
    assert response.ok, EXPECTED_CALL_TO_SUCCEED
    return json.loads(response.content)


def search_endpoints(session):
    payload = {
        "criteria": {
            "deployment_type": [
                "ENDPOINT",
                "WORKLOAD",
                "VDI",
                "AWS",
                "AZURE",
                "GCP"
            ]
        },
        "sort": [
            {
                "field": "last_contact_time",
                "order": "DESC"
            }
        ],
        "start": 1,
        "rows": 50,
        "query": ""
    }

    response = session.post(
        url=f"{session.gateway_url}/asset-registry/v1/appservices/v6/orgs/{session.get_user_org_key()}/devices/_search",
        data=json.dumps(payload)
    )
    assert response.ok, EXPECTED_CALL_TO_SUCCEED
    return json.loads(response.content)


def search_asset_groups(session):
    payload = {"sort": [{"field": "name", "order": "ASC"}], "start": 0, "rows": 50, "query": ""}

    response = session.post(
        url=f"{session.gateway_url}/asset_groups/v1/orgs/{session.get_user_org_key()}/groups/_search",
        data=json.dumps(payload)
    )
    assert response.ok, EXPECTED_CALL_TO_SUCCEED
    return json.loads(response.content)


def create_asset_group(session, policy_id=None, name=None, query=None):
    payload = {
        "member_type": "DEVICE",
        "name": name if name is not None else str(uuid.uuid4()),
        "org_key": session.get_user_org_key(),
        "description": "Test asset group created for PAS tests."
    }

    if query:
        payload["query"] = query

    if policy_id:
        payload["policy_id"] = policy_id

    response = session.post(
        url=f"{session.gateway_url}/asset_groups/v1/orgs/{session.get_user_org_key()}/groups",
        data=json.dumps(payload)
    )
    assert response.ok, EXPECTED_CALL_TO_SUCCEED
    response = json.loads(response.content)

    ASSET_GROUPS.append(response)
    return response


def add_asset_group_to_endpoint(session, asset_group_id, endpoint_id):
    add_asset_group_to_endpoint(session, asset_group_id, [endpoint_id])


def add_asset_group_to_endpoint(session, asset_group_id, endpoint_ids):
    payload = {
        "action": "CREATE",
        "external_member_ids": endpoint_ids
    }

    response = session.post(
        url=f"{session.gateway_url}/asset_groups/v1/orgs/{session.get_user_org_key()}/groups/{asset_group_id}/members",
        data=json.dumps(payload)
    )
    assert response.ok, EXPECTED_CALL_TO_SUCCEED


def remove_asset_group_from_endpoint(session, asset_group_id, endpoint_ids):
    payload = {
        "action": "REMOVE",
        "external_member_ids": endpoint_ids
    }

    response = session.post(
        url=f"{session.gateway_url}/asset_groups/v1/orgs/{session.get_user_org_key()}/groups/{asset_group_id}/members",
        data=json.dumps(payload)
    )
    assert response.ok, EXPECTED_CALL_TO_SUCCEED



def add_policy_override(session, policy_id, device_id):
    payload = {
        "action_type": "UPDATE_POLICY",
        "options": {
            "policy_id": policy_id
        },
        "device_id": [device_id]
    }

    response = session.post(
        url=f"{session.gateway_url}/appservices/v6/orgs/{session.get_user_org_key()}/device_actions",
        data=json.dumps(payload)
    )
    assert response.ok, EXPECTED_CALL_TO_SUCCEED


def delete_sensors(session, devices_to_delete=None):
    devices_to_delete = devices_to_delete if devices_to_delete is not None else DEVICES
    payload = {
        "action_type": "UNINSTALL_SENSOR",
        "device_id": [device.get('device_id') for device in devices_to_delete]
    }

    response = session.post(
        url=f"{session.gateway_url}/appservices/v6/orgs/{session.get_user_org_key()}/device_actions",
        data=json.dumps(payload)
    )
    assert response.ok, EXPECTED_CALL_TO_SUCCEED

    time.sleep(30)

    payload = {
        "action_type": "DELETE_SENSOR",
        "device_id": [device.get('device_id') for device in devices_to_delete]
    }

    response = session.post(
        url=f"{session.gateway_url}/appservices/v6/orgs/{session.get_user_org_key()}/device_actions",
        data=json.dumps(payload)
    )
    assert response.ok, EXPECTED_CALL_TO_SUCCEED


def delete_sensor(session, sensor_id):
    payload = {
        "action_type": "UNINSTALL_SENSOR",
        "device_id": [sensor_id]
    }

    response = session.post(
        url=f"{session.gateway_url}/appservices/v6/orgs/{session.get_user_org_key()}/device_actions",
        data=json.dumps(payload)
    )
    assert response.ok, EXPECTED_CALL_TO_SUCCEED

    time.sleep(30)

    payload = {
        "action_type": "DELETE_SENSOR",
        "device_id": [sensor_id]
    }

    response = session.post(
        url=f"{session.gateway_url}/appservices/v6/orgs/{session.get_user_org_key()}/device_actions",
        data=json.dumps(payload)
    )
    assert response.ok, EXPECTED_CALL_TO_SUCCEED


def delete_asset_groups(session, groups=None):
    groups_to_delete = groups if groups is not None else ASSET_GROUPS

    for asset_group in ASSET_GROUPS:
        delete_asset_group(session, asset_group.get('id'))


def delete_asset_group_with_ids(session, group_ids):
    for id in group_ids:
        delete_asset_group(session, id)


def delete_asset_group(session, group_id):
    response = session.delete(
        url=f"{session.gateway_url}/asset_groups/v1/orgs/{session.get_user_org_key()}/groups/{group_id}")
    assert response.ok, EXPECTED_CALL_TO_SUCCEED


def delete_policies(session):
    for policy in POLICIES:
        delete_policy(session, policy.get('id'))


def delete_policies_with_ids(session, policy_ids):
    for id in policy_ids:
        delete_policy(session, id)


def delete_policy(session, policy_id):
    response = session.delete(
        url=f"{session.gateway_url}/policyservice/v1/orgs/{session.get_user_org_key()}/policies/{policy_id}")
    assert response.ok, EXPECTED_CALL_TO_SUCCEED


def execute(command):
    process = subprocess.Popen(command, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, text=True)
    output, error = process.communicate(timeout=600)
    return process, output, error
